from hashlib import sha1
import logging
import requests

from . import app_settings


logger = logging.getLogger(__name__)


session = requests.Session()
session.headers.update({'User-Agent': app_settings.PWNED['USER-AGENT']})


class PwnedClient:

    def fetch_range(self, prefix):
        try:
            url = ''.join([app_settings.PWNED['ENDPOINT'], prefix])
            resp = session.get(url, timeout=app_settings.PWNED['TIMEOUT'])
            return resp.text
        except requests.exceptions.RequestException as e:
            logger.exception("Request to Pwned Password API failed. Validation skipped.")
            return ""

    def parse_range(self, raw_range):
        """
        The range of hashes are returned as suffixes only, followed
        by the number of occurrences, separated by a colon.

            <hash suffix>:<count>
            0018A45C4D1DEF81644B54AB7F969B88D65:1
            00D4F6E8FA6EECAD2A3AA415EEC418D38EC:2
            011053FD0102E94D6AE2F8B83D76FAF94F6:1
            012A7CA357541F0AC487871FEEC1891C49C:2
        """
        split_lines = [line.split(":") for line in raw_range.split('\n') if line.strip()]
        return dict((suffix.strip(), int(count)) for suffix, count in split_lines)

    def join_hash(self, prefix, suffix):
        return ''.join([prefix, suffix])

    def make_hash(self, password):
        # uppercased to match responses
        return sha1(password.encode()).hexdigest().upper()

    def split_hash(self, hashed_password):
        length = app_settings.PWNED['PREFIX_LENGTH']
        return hashed_password[:length], hashed_password[length:]

    def count_occurrences(self, password):
        hashed_password = self.make_hash(password)
        prefix, suffix = self.split_hash(hashed_password)
        raw_range = self.fetch_range(prefix)
        suffixes = self.parse_range(raw_range)
        return suffixes.get(suffix, 0)



from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

from .client import PwnedClient
from . import app_settings


@deconstructible
class PwnedValidator:
    message = _('This password is known to be weak')
    code = 'invalid'
    client = PwnedClient

    def validate(self, password, user=None):
        pwned_client = self.client()
        count = pwned_client.count_occurrences(password)
        if count >= app_settings.PWNED['OCCURRENCE_THRESHOLD']:
            raise ValidationError(self.message, code=self.code)



import pytest

from pwned.client import PwnedClient

def test_parse_range_valid():
    client = PwnedClient()
    raw_range = """
    0018A45C4D1DEF81644B54AB7F969B88D65:1
    00D4F6E8FA6EECAD2A3AA415EEC418D38EC:2
    011053FD0102E94D6AE2F8B83D76FAF94F6:1
    012A7CA357541F0AC487871FEEC1891C49C:2
    """
    expected = {
        '0018A45C4D1DEF81644B54AB7F969B88D65': 1,
        '00D4F6E8FA6EECAD2A3AA415EEC418D38EC': 2,
        '011053FD0102E94D6AE2F8B83D76FAF94F6': 1,
        '012A7CA357541F0AC487871FEEC1891C49C': 2,
    }
    assert client.parse_range(raw_range) == expected


def test_join_hash():
    client = PwnedClient()
    prefix = '00D4F'
    suffix = '6E8FA6EECAD2A3AA415EEC418D38EC'
    expected = '00D4F6E8FA6EECAD2A3AA415EEC418D38EC'
    assert client.join_hash(prefix, suffix) == expected


def test_make_hash():
    client = PwnedClient()
    password = 'My1E37pwd'
    expected = '8CEF1E00B20F463C1E48B589B03660D4E3B9EF7A'
    assert client.make_hash(password) == expected


def test_split_hash():
    client = PwnedClient()
    hashed_password = '8CEF1E00B20F463C1E48B589B03660D4E3B9EF7A'
    assert client.split_hash(hashed_password) == ('8CEF1', 'E00B20F463C1E48B589B03660D4E3B9EF7A')


def test_split_hash_prefix_length(monkeypatch):
    monkeypatch.setattr('pwned.app_settings.PWNED', {'PREFIX_LENGTH': 10})
    client = PwnedClient()
    hashed_password = '8CEF1E00B20F463C1E48B589B03660D4E3B9EF7A'
    prefix, suffix = client.split_hash(hashed_password)
    assert len(prefix) == 10


@pytest.mark.vcr
def test_count_occurrences_found():
    client = PwnedClient()
    count = client.count_occurrences('password')
    assert count == 3303003


@pytest.mark.vcr
def test_count_occurrences_none():
    client = PwnedClient()
    count = client.count_occurrences('8CEF1E00B20F463C1E48B589B03660D4E3B9EF7A')
    assert count == 0



import os
from setuptools import find_packages, setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-pwned-validator',
    version='0.0.3',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='A Pwned Passwords validator for Django',
    long_description=open('README.rst').read(),
    url='https://github.com/craigloftus/django-pwned-validator',
    author='Craig Loftus',
    author_email='craigloftus@gmail.com',
    python_requires='>=3.5',
    install_requires=[
        'requests',
    ],
    extras_require={
        'test': ['django<2.1', 'pytest', 'pytest-cov', 'pytest-django', 'pytest-vcr',],
    },
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
)



from django.conf import settings


DEFAULTS = {
    'ENDPOINT': 'https://api.pwnedpasswords.com/range/',
    'TIMEOUT': 2, # The default is conservative, it should be <20ms typically, 500ms if uncached
    'PREFIX_LENGTH': 5,
    'OCCURRENCE_THRESHOLD': 1, # How many occurrences is too many
    'USER-AGENT': 'github.com/craigloftus/django-pwned-validator',
}


PWNED = getattr(settings, 'PWNED', DEFAULTS)
